import enum


def budget_desc_to_ledger_account(budget_desc: str) -> str:
    """Convert a workday category to an AccountCategory enum."""
    lookup = {
        "BYU Grants: Benefits": "5900:Benefits",
        "BYU Grants: Indirect Costs": "6400:Indirect Costs",
        "BYU Grants: Materials and Supplies": "6100:Materials and Supplies",
        "BYU Grants: Salaries and Wages": "5000:Salaries and Wages",
        "BYU Grants: Student Aid": "6300:Student Aid",
        "BYU Grants: Travel": "7000:Travel",
        "BYU Grants: Unallocated": "9999:Unallocated",
        "BYU Grants: Capital Equipment": "6900:Capital Equipment",
        "BYU Grants: Contract Services": "6450:Contract Services",
        "BYU Grants: Transfers": "9250:Transfers",
    }
    if budget_desc in lookup:
        return lookup[budget_desc]
    else:
        raise ValueError(f"Unknown category: {budget_desc}. Please update the lookup.")


# Mapping of transaction ledger accounts to budget category ledger accounts.
# Some transaction ledger accounts (e.g. 6705) don't directly match a budget
# category, so we map them to the appropriate one.
LEDGER_ACCOUNT_MAPPING = {
    6700: 6100,  # -> Materials and Supplies
}


class BudgetType(enum.Enum):
    """Enum for budget types."""

    ACTUALS = 0
    BUDGET = 1
    ACTUALS_YTD = 2
    ACTUALS_PREV = 3
    COMMITTED = 4


class AccountType(enum.Enum):
    """Enum for account types."""

    GR = "GR"
    AC = "AC"
    GF = "GF"



import argparse
from datetime import datetime
import pathlib

from .xlsx import xlsx_merge
from .config import DEFAULT_CHROME_PORT, set_new_chrome_tabs, set_verbose
from .account import Account
from .chrome import get_chrome_webdriver, run_chrome
from .yutils import error


def run_chrome_command(args):
    run_chrome(port=args.port)


def report_actuals_command(args):
    if args.verbose:
        set_verbose(True)
    if args.tabs:
        set_new_chrome_tabs(True)

    dr = get_chrome_webdriver(remote_debugging_port=args.port)

    account = Account(
        args.account,
        args.year,
        args.transactions_xlsx_out,
        args.payroll_xlsx_out,
        args.journals_xlsx_out,
    )
    account.get_workday_data(dr)

    account.to_excel(args.xlsx_out)


def xlsx_merge_command(args):
    if not args.input_files or len(args.input_files) < 2:
        error("You must specify at least two input files to merge.")
    xlsx_merge(args.input_files, args.output_file)


def main():
    parser = argparse.ArgumentParser(description="YAccounts Workday CLI")
    subparsers = parser.add_subparsers(dest="command", required=True)

    #################### Run Chrome Command ####################
    launch_chrome_parser = subparsers.add_parser(
        "run-chrome", help="Launch Chrome browser"
    )
    launch_chrome_parser.add_argument(
        "--port",
        type=int,
        default=DEFAULT_CHROME_PORT,
        help=f"Port to use for Chrome remote debugging (default: {DEFAULT_CHROME_PORT})",
    )
    launch_chrome_parser.set_defaults(func=run_chrome_command)

    #################### Report Actuals Command ####################
    report_actuals_parser = subparsers.add_parser(
        "report-actuals", help="Run the budget and actuals report"
    )
    report_actuals_parser.add_argument(
        "account",
        type=str,
        help="Workday account ID (e.g., GR00227)",
    )
    report_actuals_parser.add_argument(
        "--year",
        type=int,
        default=datetime.now().year,
        help="Year for the budget report (default: current year)",
    )
    report_actuals_parser.add_argument(
        "--xlsx-out",
        type=pathlib.Path,
        help="Path to save the output Excel file (default: ./<account>_<year>.xlsx)",
    )
    report_actuals_parser.add_argument(
        "--transactions-xlsx-out",
        type=pathlib.Path,
        help="Path to save the transaction data as an Excel file. Only applicable to GR accounts. (optional, default: None)",
    )
    report_actuals_parser.add_argument(
        "--journals-xlsx-out",
        type=pathlib.Path,
        help="Path to save the journal data as an Excel file. Only applicable to non-GR accounts. (optional, default: None)",
    )
    report_actuals_parser.add_argument(
        "--payroll-xlsx-out",
        type=pathlib.Path,
        help="Path to save the payroll data as an Excel file. Only applicable to non-GR accounts. (optional, default: None)",
    )
    report_actuals_parser.add_argument(
        "--port",
        type=int,
        default=DEFAULT_CHROME_PORT,
        help=f"Port Chrome is running on (default: {DEFAULT_CHROME_PORT})",
    )
    report_actuals_parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Enable verbose output",
    )
    report_actuals_parser.add_argument(
        "--tabs",
        action="store_true",
        help="Open new Chrome tabs for each operation (default: False)",
    )
    report_actuals_parser.set_defaults(func=report_actuals_command)

    #################### XLSX Merge Command #####################
    xlsx_merge_parser = subparsers.add_parser(
        "xlsx-merge", help="Merge multiple Excel files into one"
    )
    xlsx_merge_parser.add_argument(
        "output_file",
        type=pathlib.Path,
        help="Path to the output Excel file where merged data will be saved",
    )
    xlsx_merge_parser.add_argument(
        "input_files",
        nargs="+",
        type=pathlib.Path,
        help="Paths to the input Excel files to merge",
    )
    xlsx_merge_parser.set_defaults(func=xlsx_merge_command)

    args = parser.parse_args()

    # Automatically call the right function!
    args.func(args)



from collections import defaultdict
import re
import time

import pandas as pd
from bs4 import BeautifulSoup
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys

from .config import WAIT_TIME_BASE, WAIT_TIME_TABLE, WAIT_TIME_TABLE_PAGES
from .budget_categories import BudgetType, budget_desc_to_ledger_account
from .yutils import print_progress, workday_str_amount_to_float


class TableExtractionError(Exception):
    """Custom exception for errors during table extraction."""


def extract_payroll_data_from_table(dr):
    return _extract_df_from_simple_table(dr)


def extract_journal_data_from_table(dr):
    return _extract_df_from_simple_table(dr)


def extract_transaction_data_from_table(dr):
    return _extract_df_from_simple_table(dr)


def extract_actuals_from_grant_table(dr, expected_month):
    """Extracts actuals from the data table.
    expected_month: (datetime) is used to verify that the correct report is loaded.
    """
    # Wait for the data table to be present and visible
    table_div = WebDriverWait(dr, WAIT_TIME_TABLE).until(
        EC.visibility_of_element_located((By.CSS_SELECTOR, "div.wd-DataGrid"))
    )

    # Expand grant row
    _expand_all_rows(dr, table_div, min_rows=0, max_rows=1)

    table_html = table_div.get_attribute("outerHTML")
    soup = BeautifulSoup(table_html, "html.parser")

    # Find the first <table> inside the data grid (header table)
    headers = _extract_headers_from_grant_table(soup)

    # Ensure data for correct month was returned
    if not headers[2].startswith(f"{expected_month.strftime('%Y - %b')}"):
        raise TableExtractionError

    # Find the first <table> with class containing 'grid-body-row' (data rows)
    body_table = soup.find("table", class_=lambda x: x and "grid-body-row" in x)
    extracted_data = defaultdict(dict)

    for row in body_table.find_all("tr", attrs={"data-automation-id": "gridrow"})[2:]:
        # Skip the first two rows ("Object Class" and Award Name)
        cells = row.find_all("td", attrs={"data-automation-id": "gridCell"})
        cells = [cell.text.strip() for cell in cells]

        if cells[0] == "Total":
            continue

        category = budget_desc_to_ledger_account(cells[0])

        extracted_data[category][BudgetType.ACTUALS] = workday_str_amount_to_float(
            cells[3]
        )
        extracted_data[category][BudgetType.BUDGET] = workday_str_amount_to_float(
            cells[2]
        )
        extracted_data[category][BudgetType.ACTUALS_YTD] = workday_str_amount_to_float(
            cells[4]
        )
        extracted_data[category][BudgetType.COMMITTED] = workday_str_amount_to_float(
            cells[5]
        )
        extracted_data[category][BudgetType.ACTUALS_PREV] = (
            extracted_data[category][BudgetType.ACTUALS_YTD]
            - extracted_data[category][BudgetType.ACTUALS]
        )

    return extracted_data


def extract_actuals_from_activity_table(dr, expected_month):
    """Extracts actuals from the data table.
    expected_month: (datetime) is used to verify that the correct report is loaded.
    """
    # Wait for the data table to be present and visible
    table_div = WebDriverWait(dr, WAIT_TIME_TABLE).until(
        EC.visibility_of_element_located((By.CSS_SELECTOR, "div.wd-DataGrid"))
    )

    # Expand grant row
    _expand_all_rows(dr, table_div, min_rows=1, max_rows=1)

    # Expand revenue/income/transfer rows
    _expand_all_rows(dr, table_div, min_rows=1, max_rows=3)

    table_html = table_div.get_attribute("outerHTML")
    soup = BeautifulSoup(table_html, "html.parser")

    headers = _extract_headers_from_grant_table(soup)

    # Ensure data for correct month was returned
    if not headers[1].startswith(f"{expected_month.strftime('%b %Y')}"):
        raise TableExtractionError

    # Find the first <table> with class containing 'grid-body-row' (data rows)
    body_table = soup.find("table", class_=lambda x: x and "grid-body-row" in x)

    extracted_data = defaultdict(dict)

    for row in body_table.find_all("tr", attrs={"data-automation-id": "gridrow"}):
        cells = row.find_all("td", attrs={"data-automation-id": "gridCell"})
        cells = [cell.text.strip() for cell in cells]

        # Skip the header rows
        if (
            cells[0] == "Gift"
            or cells[0].startswith("AC")
            or cells[0].startswith("GF")
            or re.match(r"\d - ", cells[0])
            or cells[0] == "(Blank)"
        ):
            continue

        assert re.match(r"\d\d\d\d:", cells[0]), f"Unexpected row format: {cells[0]}"

        extracted_data[cells[0]][BudgetType.ACTUALS] = workday_str_amount_to_float(
            cells[2]
        )
        extracted_data[cells[0]][BudgetType.BUDGET] = workday_str_amount_to_float(
            cells[1]
        )
        extracted_data[cells[0]][BudgetType.ACTUALS_YTD] = workday_str_amount_to_float(
            cells[3]
        )
        extracted_data[cells[0]][BudgetType.COMMITTED] = workday_str_amount_to_float(
            cells[4]
        )
        extracted_data[cells[0]][BudgetType.ACTUALS_PREV] = (
            extracted_data[cells[0]][BudgetType.ACTUALS_YTD]
            - extracted_data[cells[0]][BudgetType.ACTUALS]
        )

    return extracted_data


################################################################################
############################# Internal Helpers #################################
################################################################################
def _extract_headers_from_grant_table(soup):
    # Find the first <table> inside the data grid (header table)
    header_table = soup.find("table", class_="grid")
    headers = []
    if header_table:
        header_row = header_table.find("tr", class_="grid-head-row")
        if header_row:
            for th in header_row.find_all("td", attrs={"role": "columnheader"}):
                label = th.find("span", attrs={"data-automation-id": "columnLabel-0"})
                headers.append(label.text.strip() if label else th.text.strip())
    return headers


def _extract_df_from_simple_table(dr):
    """Extracts a simple table from the current page and returns it as a DataFrame."""

    # Find the table and get the headers
    table_div = dr.find_element(By.CSS_SELECTOR, 'div[data-testid="tableWrapper"]')
    table_html = table_div.get_attribute("outerHTML")
    soup = BeautifulSoup(table_html, "html.parser")

    headers = []
    header_row = soup.find("tr", attrs={"data-automation-id": "tableHeaderRow"})
    for th in header_row.find_all("th", attrs={"role": "columnheader"}):
        headers.append(th.find("span").get_text(strip=True))

    # Create empty dataframe with headers
    df = pd.DataFrame(columns=headers)

    row_count_elem = WebDriverWait(dr, WAIT_TIME_BASE).until(
        EC.presence_of_element_located(
            (By.CSS_SELECTOR, '[data-automation-id="rowCountLabel"]')
        )
    )
    num_rows = row_count_elem.text.strip()
    m = re.match(r"(\d+) item", num_rows)
    assert m, f"Unexpected row count format: {num_rows}"
    num_rows = int(m.group(1))
    multiple_pages = num_rows > 30

    # Change the max rows to 30 to avoid hidden scrollable rows
    if multiple_pages:
        _set_max_rows_to_30(dr)

    while True:
        # Get the current active page number
        if multiple_pages:
            current_page_button = WebDriverWait(dr, 10).until(
                EC.presence_of_element_located(
                    (By.CSS_SELECTOR, 'button[aria-current="page"]')
                )
            )
            current_page = int(current_page_button.text.strip())
            print_progress(f"Currently on page {current_page}.")

        # Find the number of items on this page
        if multiple_pages:
            pagination_details_elem = WebDriverWait(dr, 10).until(
                EC.presence_of_element_located(
                    (
                        By.CSS_SELECTOR,
                        '[data-automation-id="paginationAdditionalDetails"]',
                    )
                )
            )
            pagination_details = pagination_details_elem.text.strip()
            matches = re.match(r"^(\d+)-(\d+) of \d+", pagination_details)
            assert matches, f"Unexpected pagination format: {pagination_details}"
            num_items_on_page = int(matches.group(2)) - int(matches.group(1)) + 1
        else:
            num_items_on_page = num_rows

        # Wait until the row data has loaded
        while True:
            table_div = dr.find_element(
                By.CSS_SELECTOR, 'div[data-testid="tableWrapper"]'
            )
            table_html = table_div.get_attribute("outerHTML")
            soup = BeautifulSoup(table_html, "html.parser")
            data_rows = soup.find_all("tr", attrs={"data-automation-id": "row"})
            if len(data_rows) == num_items_on_page:
                break
            time.sleep(WAIT_TIME_TABLE_PAGES)

        # Find all data rows, and add them to the DataFrame
        data_rows = soup.find_all("tr", attrs={"data-automation-id": "row"})
        print_progress(f"Found {len(data_rows)} rows on this page.")
        for row in data_rows:
            row_data = []
            cells = row.find_all("td", attrs={"data-automation-id": "cell"})
            for cell in cells:
                text = cell.get_text(strip=True)
                row_data.append(text)

            if row_data:
                df.loc[len(df)] = row_data

        if not multiple_pages:
            print_progress("Single page table extracted.")
            break

        # Find the 'Next' button
        next_button = dr.find_element(
            By.CSS_SELECTOR, 'button[data-automation-id="navigateNextPage"]'
        )

        # Check if it's disabled
        if next_button.get_attribute("aria-disabled") == "true":
            print_progress("Reached last page.")
            break
        next_button.click()

        print_progress(f"Clicked to go to page {current_page + 1}.")

        # Wait for page number to update
        WebDriverWait(dr, WAIT_TIME_BASE).until(
            EC.text_to_be_present_in_element(
                (By.CSS_SELECTOR, 'button[aria-current="page"]'),
                str(current_page + 1),
            )
        )

    assert len(df) == num_rows, f"Expected {num_rows} rows, found {len(df)}"
    return df


def _expand_all_rows(dr, table_div, min_rows=None, max_rows=None):
    expand_buttons = table_div.find_elements(
        By.CSS_SELECTOR, 'div[data-automation-id="expand"][aria-expanded="false"]'
    )

    if min_rows is not None:
        assert (
            len(expand_buttons) >= min_rows
        ), f"Expected at least {min_rows} expandable rows, found {len(expand_buttons)}"
    if max_rows is not None:
        assert (
            len(expand_buttons) <= max_rows
        ), f"Expected at most {max_rows} expandable rows, found {len(expand_buttons)}"

    for expand_button in expand_buttons:
        dr.execute_script(
            "arguments[0].scrollIntoView({block: 'center'});", expand_button
        )
        expand_button.click()
        time.sleep(3)  # Give time for expansion


def _extract_text(cell):
    # Try to get button value if present (for numbers)
    btn = cell.find("button", attrs={"data-automation-id": "drillDownNumberLabel"})
    if btn:
        return btn["aria-label"].strip()
    else:
        # Try to get text from gwt-Label or fallback to cell text
        label = cell.find("div", class_="gwt-Label")
        if label:
            return label.text.strip()
        else:
            return cell.text.strip()


def _set_max_rows_to_30(driver):
    # 1. Find the dropdown input
    dropdown_input = WebDriverWait(driver, WAIT_TIME_BASE).until(
        EC.element_to_be_clickable(
            (By.CSS_SELECTOR, 'input[data-testid="showAllRowsDropdownInput"]')
        )
    )

    # 2. Click to open the dropdown
    dropdown_input.click()

    # 3. Clear existing text (optional)
    dropdown_input.clear()

    # 4. Type "30"
    dropdown_input.send_keys("30")

    # 5. Press Enter to select
    dropdown_input.send_keys(Keys.ENTER)

    print_progress("Set max rows per page to 30.")



DEFAULT_CHROME_PORT = 9222
DEVELOPER_NAME = "Jeff Goeders"

VERBOSE = False

DEBUG_WITH_CACHING = False

END_MONTH = None

WAIT_TIME_BASE = 20
WAIT_TIME_TABLE = 120
WAIT_TIME_GRANT = 5
WAIT_TIME_PERIOD = 5
WAIT_TIME_MENU_BUTTON = 5
WAIT_TIME_TABLE_PAGES = 0.25


def set_verbose(verbose: bool):
    """Set the verbosity level for logging."""
    global VERBOSE
    VERBOSE = verbose


def get_verbose():
    """Get the current verbosity level."""
    return VERBOSE


NEW_CHROME_TABS = False


def set_new_chrome_tabs(new_tabs: bool):
    """Set whether to open new Chrome tabs for each operation."""
    global NEW_CHROME_TABS
    NEW_CHROME_TABS = new_tabs


def get_new_chrome_tabs():
    """Get whether to open new Chrome tabs for each operation."""
    return NEW_CHROME_TABS



import copy
from openpyxl import load_workbook, Workbook
from openpyxl.cell.cell import MergedCell
from openpyxl.utils import get_column_letter

from .yutils import TermColors, print_color


def xlsx_merge(input_files, output_file):
    """Merge multiple Excel files into one, preserving styles and merged cells."""
    print_color(
        TermColors.GREEN,
        f"Merging {[str(s) for s in input_files]} files into {output_file}...",
    )
    output_wb = Workbook()
    output_ws = output_wb.active
    output_ws.title = "Merged"
    current_row_idx = 1

    output_ws.sheet_properties.outlinePr.summaryBelow = False
    output_ws.sheet_properties.outlinePr.applyStyles = True

    for idx, file_path in enumerate(input_files):
        wb = load_workbook(file_path)
        ws = wb.active

        if idx == 0:
            for col_letter, dim in ws.column_dimensions.items():
                output_ws.column_dimensions[col_letter].width = dim.width

        row_offset = current_row_idx - 1

        # Track merged ranges for post-formatting
        merged_ranges = []

        for merged_range in ws.merged_cells.ranges:
            min_col, min_row, max_col, max_row = merged_range.bounds
            new_range = (
                f"{get_column_letter(min_col)}{min_row + row_offset}:"
                f"{get_column_letter(max_col)}{max_row + row_offset}"
            )
            output_ws.merge_cells(new_range)
            merged_ranges.append(((min_col, min_row, max_col, max_row), row_offset))

        for src_row in ws.iter_rows():
            src_row_idx = src_row[0].row
            dst_row_idx = current_row_idx

            for cell in src_row:
                if isinstance(cell, MergedCell):
                    continue

                dst_cell = output_ws.cell(
                    row=dst_row_idx, column=cell.column, value=cell.value
                )

                if cell.has_style:
                    dst_cell.font = copy.copy(cell.font)
                    dst_cell.border = copy.copy(cell.border)
                    dst_cell.fill = copy.copy(cell.fill)
                    dst_cell.number_format = copy.copy(cell.number_format)
                    dst_cell.protection = copy.copy(cell.protection)
                    dst_cell.alignment = copy.copy(cell.alignment)

            # Grouping & collapsed state
            src_dim = ws.row_dimensions[src_row_idx]
            dst_dim = output_ws.row_dimensions[dst_row_idx]
            dst_dim.outlineLevel = src_dim.outlineLevel
            dst_dim.hidden = src_dim.hidden

            current_row_idx += 1

        # Reapply styles to all cells in merged regions
        for (min_col, min_row, max_col, max_row), offset in merged_ranges:
            src_cell = ws.cell(row=min_row, column=min_col)
            for row in range(min_row + offset, max_row + offset + 1):
                for col in range(min_col, max_col + 1):
                    cell = output_ws.cell(row=row, column=col)
                    if src_cell.has_style:
                        cell.font = copy.copy(src_cell.font)
                        cell.border = copy.copy(src_cell.border)
                        cell.fill = copy.copy(src_cell.fill)
                        cell.number_format = copy.copy(src_cell.number_format)
                        cell.protection = copy.copy(src_cell.protection)
                        cell.alignment = copy.copy(src_cell.alignment)

        if idx < len(input_files) - 1:
            output_ws.row_dimensions[current_row_idx].height = 40  # Tall gap row
            current_row_idx += 1

    output_wb.save(output_file)



import calendar
from dataclasses import dataclass
from datetime import datetime
import enum
import pathlib

from openpyxl import load_workbook
from openpyxl.styles import PatternFill, Font, Alignment, Border, Side
from openpyxl.utils import get_column_letter
from openpyxl.worksheet.dimensions import SheetFormatProperties
import pandas as pd

from .budget_categories import AccountType, BudgetType, LEDGER_ACCOUNT_MAPPING
from .config import DEBUG_WITH_CACHING, DEVELOPER_NAME, END_MONTH
from .run_report import (
    run_activity_report,
    run_gift_report,
    run_grant_report,
    run_journal_report,
    run_payroll_report,
    run_transactions_report,
)
from .yutils import TermColors, print_color, warn, workday_str_amount_to_float, error

ROW_INDENT_STR = " ▸ "

MONTH_COLS = calendar.month_abbr[1:]


@dataclass(frozen=True)
class _ColumnHeader:
    col_name: str
    index: int


class _ColumnHeaders(enum.Enum):
    """Enum for column indices in the DataFrame."""

    CATEGORY = _ColumnHeader("Category", 1)
    BUDGET = _ColumnHeader("Budget", 2)
    PREV_ACTUALS = _ColumnHeader("Prev Actuals", 3)
    JAN = _ColumnHeader("Jan", 4)
    FEB = _ColumnHeader("Feb", 5)
    MAR = _ColumnHeader("Mar", 6)
    APR = _ColumnHeader("Apr", 7)
    MAY = _ColumnHeader("May", 8)
    JUN = _ColumnHeader("Jun", 9)
    JUL = _ColumnHeader("Jul", 10)
    AUG = _ColumnHeader("Aug", 11)
    SEP = _ColumnHeader("Sep", 12)
    OCT = _ColumnHeader("Oct", 13)
    NOV = _ColumnHeader("Nov", 14)
    DEC = _ColumnHeader("Dec", 15)
    ACTUALS_YTD = _ColumnHeader("YTD", 16)
    COMMITTED = _ColumnHeader("Pending Commitments", 17)
    BALANCE = _ColumnHeader("Balance", 18)


def _get_column_headers():
    """Get the column headers as a list of strings."""
    return [col.value.col_name for col in _ColumnHeaders]


class Account:
    def __init__(
        self,
        account_code,
        year,
        transactions_xlsx_out_path=None,
        payroll_xlsx_out_path=None,
        journals_xlsx_out_path=None,
    ):

        self.account_code = account_code
        self.year = year
        self.transactions_xlsx_out_path = transactions_xlsx_out_path
        self.payroll_xlsx_out_path = payroll_xlsx_out_path
        self.journals_xlsx_out_path = journals_xlsx_out_path

        try:
            self.account_type = AccountType[account_code[:2].upper()]
        except KeyError:
            error(
                f"Account code {self.account_code} does not start with 'GR', 'AC' or 'GF'. "
                "Other account codes are not supported yet."
            )

        if self.account_type == AccountType.GR:
            assert self.payroll_xlsx_out_path is None, (
                "Payroll data is not applicable for GR accounts. "
                "Please set payroll_xlsx_out_path to None."
            )
            assert self.journals_xlsx_out_path is None, (
                "Journal data is not applicable for GR accounts. "
                "Please set journals_xlsx_out_path to None."
            )
        else:
            assert self.transactions_xlsx_out_path is None, (
                "Transaction data is not applicable for non-GR accounts. "
                "Please set transactions_xlsx_out_path to None."
            )

        self._df = pd.DataFrame(columns=_get_column_headers())

        self._df_transactions = None
        self._df_journals = None
        self._df_payroll = None

    def get_workday_data(self, dr):
        """Get all workday data for the account."""

        # This will run the workday reports.
        # First the annual reports are run (transactaions for GR accounts, payroll for non-GR accounts),
        # Then the month-by-month reports are run to get monthly balances (and monthly journal transactions for non-GR accounts).
        # This saves all the workday data into pandas DataFrames (self._df* properties).
        if self.uses_transaction_report_data():
            self._get_transactions(dr)
        else:
            self._get_payroll_data(dr)
        self._get_monthly_data(dr)

        # Next, the data is processed and merged into the main DataFrame (self._df).
        self._finalize_workday_data()
        self._add_transaction_or_journal_data()

        if not self.uses_transaction_report_data():
            self._add_payroll_data()

    def uses_transaction_report_data(self):
        """Check if the account uses data from the new grant transaction report (only available to GR accounts)."""
        return self.account_type == AccountType.GR

    ############################################################################
    ##################### Methods to get data from workday #####################
    ############################################################################

    def _get_transactions(self, dr):
        """Get transactions from Workday."""

        cache_filename = pathlib.Path(
            f"cache/transactions_{self.account_code}_{self.year}.pkl"
        )
        if DEBUG_WITH_CACHING and cache_filename.exists():
            print_color(
                TermColors.BLUE,
                f"Loading transactions from cache: {cache_filename}",
            )
            self._df_transactions = pd.read_pickle(cache_filename)
            return

        self._df_transactions = run_transactions_report(
            dr, self.account_code, self.year
        )
        if DEBUG_WITH_CACHING:
            cache_filename.parent.mkdir(parents=True, exist_ok=True)
            self._df_transactions.to_pickle(cache_filename)
            self._df_transactions.to_csv(
                cache_filename.with_suffix(".csv"), index=False
            )

    def _get_payroll_data(self, dr):
        """Get payroll data from Workday."""

        cache_filename = pathlib.Path(
            f"cache/payroll_{self.account_code}_{self.year}.pkl"
        )
        if DEBUG_WITH_CACHING and cache_filename.exists():
            print_color(
                TermColors.BLUE,
                f"Loading payroll data from cache: {cache_filename}",
            )
            self._df_payroll = pd.read_pickle(cache_filename)
            return

        self._df_payroll = run_payroll_report(dr, self.account_code, self.year)
        if DEBUG_WITH_CACHING:
            cache_filename.parent.mkdir(parents=True, exist_ok=True)
            self._df_payroll.to_pickle(cache_filename)
            self._df_payroll.to_csv(cache_filename.with_suffix(".csv"), index=False)

    def _get_monthly_data(self, dr):
        """Get monthly data from Workday."""

        cache_filename_monthly = pathlib.Path(
            f"cache/monthly_{self.account_code}_{self.year}.pkl"
        )
        cache_filename_journals = pathlib.Path(
            f"cache/journals_{self.account_code}_{self.year}.pkl"
        )

        if (
            DEBUG_WITH_CACHING
            and cache_filename_monthly.exists()
            and cache_filename_journals.exists()
        ):
            print_color(
                TermColors.BLUE,
                f"Loading monthly data from cache: {cache_filename_monthly}",
            )
            self._df = pd.read_pickle(cache_filename_monthly)
            print_color(
                TermColors.BLUE,
                f"Loading journal data from cache: {cache_filename_journals}",
            )
            self._df_journals = pd.read_pickle(cache_filename_journals)
            return

        end_month = 12 if self.year != datetime.now().year else datetime.now().month
        if END_MONTH is not None:
            end_month = END_MONTH
        month_year_list = [
            datetime(self.year, month, 1) for month in range(1, end_month + 1)
        ]

        for month in month_year_list:

            if self.account_code.startswith("GR"):
                data = run_grant_report(dr, self.account_code, month)
            elif self.account_code.startswith("AC"):
                data = run_activity_report(dr, self.account_code, month)
            elif self.account_code.startswith("GF"):
                data = run_gift_report(dr, self.account_code, month)
            if self.account_type != AccountType.GR:
                journal_data_df = run_journal_report(dr, self.account_code, month)

                if self._df_journals is None:
                    self._df_journals = journal_data_df.copy()
                else:
                    self._df_journals = pd.concat(
                        [self._df_journals, journal_data_df], ignore_index=True
                    )

            for category in data:
                # If the actuals_df does not have a row for this category, add it
                if (
                    category
                    not in self._df[_ColumnHeaders.CATEGORY.value.col_name].values
                ):
                    new_row = {_ColumnHeaders.CATEGORY.value.col_name: category}
                    self._df = pd.concat(
                        [self._df, pd.DataFrame([new_row])], ignore_index=True
                    )

                # Update the actuals_df for this category and month
                self._df.loc[
                    self._df[_ColumnHeaders.CATEGORY.value.col_name] == category,
                    month.strftime("%b"),
                ] = data[category][BudgetType.ACTUALS]

                # If January, update previous actuals
                if month.month == 1:
                    self._df.loc[
                        self._df[_ColumnHeaders.CATEGORY.value.col_name] == category,
                        _ColumnHeaders.PREV_ACTUALS.value.col_name,
                    ] = data[category][BudgetType.ACTUALS_PREV]

                # If month is last month of loop, update budget and committed
                if month.month == end_month:
                    self._df.loc[
                        self._df[_ColumnHeaders.CATEGORY.value.col_name] == category,
                        _ColumnHeaders.BUDGET.value.col_name,
                    ] = data[category][BudgetType.BUDGET]
                    self._df.loc[
                        self._df[_ColumnHeaders.CATEGORY.value.col_name] == category,
                        _ColumnHeaders.COMMITTED.value.col_name,
                    ] = data[category][BudgetType.COMMITTED]

        if DEBUG_WITH_CACHING:
            cache_filename_monthly.parent.mkdir(parents=True, exist_ok=True)
            self._df.to_pickle(cache_filename_monthly)
            self._df.to_csv(cache_filename_monthly.with_suffix(".csv"), index=False)
            if self._df_journals is not None:
                self._df_journals.to_pickle(cache_filename_journals)
                self._df_journals.to_csv(
                    cache_filename_journals.with_suffix(".csv"), index=False
                )

    ############################################################################
    ##################### Methods to process workday data ######################
    ############################################################################

    def _finalize_workday_data(self):
        """Finalize and cleanup all workday data"""
        #################### Process actuals (self._df) ####################
        # Convert all columns except 'Category' to float
        for col in self._df.columns:
            if col != _ColumnHeaders.CATEGORY.value.col_name:
                self._df[col] = self._df[col].apply(
                    lambda x: float(x) if x != "" else ""
                )

        # Replace NaN values with 0
        self._df = self._df.fillna(0.0)

        # Sort by Category
        self._df = self._df.sort_values(
            by=_ColumnHeaders.CATEGORY.value.col_name
        ).reset_index(drop=True)

        # YTD column is the sum of all months
        self._df[_ColumnHeaders.ACTUALS_YTD.value.col_name] = self._df[
            [datetime(1900, month, 1).strftime("%b") for month in range(1, 13)]
        ].sum(axis=1)

        # Balance is Budget - Prev Actuals - YTD - Committed
        self._df[_ColumnHeaders.BALANCE.value.col_name] = (
            self._df[_ColumnHeaders.BUDGET.value.col_name]
            - self._df[_ColumnHeaders.PREV_ACTUALS.value.col_name]
            - self._df[_ColumnHeaders.ACTUALS_YTD.value.col_name]
            - self._df[_ColumnHeaders.COMMITTED.value.col_name]
        )

        # Add a totals row
        total_row = self._df.sum(numeric_only=True)
        total_row[_ColumnHeaders.CATEGORY.value.col_name] = "Total"
        total_row[_ColumnHeaders.BUDGET.value.col_name] = total_row[
            _ColumnHeaders.BUDGET.value.col_name
        ]
        total_row[_ColumnHeaders.PREV_ACTUALS.value.col_name] = total_row[
            _ColumnHeaders.PREV_ACTUALS.value.col_name
        ]
        total_row[_ColumnHeaders.BALANCE.value.col_name] = total_row[
            _ColumnHeaders.BALANCE.value.col_name
        ]
        total_row[_ColumnHeaders.ACTUALS_YTD.value.col_name] = total_row[
            _ColumnHeaders.ACTUALS_YTD.value.col_name
        ]
        self._df = pd.concat([self._df, pd.DataFrame([total_row])], ignore_index=True)

        if self.uses_transaction_report_data():
            self._cleanup_transaction_data()
        else:
            self._cleanup_journal_data()

    def _cleanup_transaction_data(self):
        """Cleanup self._df_transactions"""

        if self._df_transactions is None or self._df_transactions.empty:
            return

        # Remove any rows where 'Accounting Date' is empty (these are total rows and not useful)
        self._df_transactions = self._df_transactions[
            self._df_transactions["Accounting Date"].notna()
            & (self._df_transactions["Accounting Date"].str.strip() != "")
        ].copy()

        # Export transactions to Excel if specified
        if self.transactions_xlsx_out_path:
            self._df_transactions.to_excel(
                self.transactions_xlsx_out_path, index=False, engine="openpyxl"
            )

        # Split 'Ledger Account' on ':' and convert to int
        self._df_transactions["Ledger Account"] = (
            self._df_transactions["Ledger Account"].str.split(":").str[0].astype(int)
        )

        # Remap ledger accounts that don't directly match a budget category
        self._df_transactions["Ledger Account"] = self._df_transactions[
            "Ledger Account"
        ].replace(LEDGER_ACCOUNT_MAPPING)

    def _cleanup_journal_data(self):
        # Remove any rows where 'Accounting Date' is empty (these are total rows and not useful)
        self._df_journals = self._df_journals[
            self._df_journals["Accounting Date"].notna()
            & (self._df_journals["Accounting Date"].str.strip() != "")
        ].copy()

        # Convert 'Ledger Account by Identifier' to int
        self._df_journals["Ledger Account by Identifier"] = self._df_journals[
            "Ledger Account by Identifier"
        ].astype(int)

    def _add_transaction_or_journal_data(self):
        df_data = (
            self._df_transactions
            if self.uses_transaction_report_data()
            else self._df_journals
        )

        if df_data is None or df_data.empty:
            return

        account_col_name = "Ledger Account"
        if not self.uses_transaction_report_data():
            # The column name is slightly different in the journal data
            account_col_name += " by Identifier"

        handled_ledger_accounts = set()
        # handled_ledger_accounts.add(6400)  # F&A individual items
        handled_ledger_accounts.add(4200)  # Award revenue

        if not self.uses_transaction_report_data():
            # If we aren't using transaction data, then we are using separate journal/payroll data,
            # so remove the Salaries and Wages category from the journal data, as this will be in the
            # payroll data.
            handled_ledger_accounts.add(5000)

        # Loop through rows in the actuals DataFrame
        for index, row in self._df.iterrows():
            # Skip sub-rows (those starting with ROW_INDENT_STR)
            if (
                row[_ColumnHeaders.CATEGORY.value.col_name].startswith(ROW_INDENT_STR)
                or row[_ColumnHeaders.CATEGORY.value.col_name] == "Total"
            ):
                continue

            # Get the ledger account number from the row
            ledger_account = int(row[_ColumnHeaders.CATEGORY.value.col_name][:4])

            # Skip these
            if ledger_account in handled_ledger_accounts:
                continue
            handled_ledger_accounts.add(ledger_account)

            # Find transaction entries for this ledger account
            transactions = df_data[
                (df_data[account_col_name] == int(ledger_account))
            ].copy()

            # If using journals, filter out ADM (administration) transactions
            if not self.uses_transaction_report_data():
                transactions = transactions[
                    ~transactions["Cost Center Name"].str.startswith("[ADM]")
                ]
            if transactions.empty:
                continue

            # Clean the 'Amount' column
            transactions["Amount"] = transactions["Amount"].apply(
                workday_str_amount_to_float
            )

            # Ensure Accounting Date is datetime
            transactions["Accounting Date"] = pd.to_datetime(
                transactions["Accounting Date"], errors="coerce"
            )

            # Add a 'Month' column with abbreviation (e.g., 'Jan', 'Feb'), and 'month_idx'
            transactions["Month"] = transactions["Accounting Date"].dt.month.apply(
                lambda m: calendar.month_abbr[m]
            )
            transactions["month_idx"] = transactions["Accounting Date"].dt.month

            # Create month columns and initialize them to None or 0.0
            for month in MONTH_COLS:
                transactions[month] = None

            def assign_amount(row):
                row[row["Month"]] = row["Amount"]
                return row

            transactions = transactions.apply(assign_amount, axis=1)

            # If ledger account is 5000, then this is employee salary data, and we
            # want to group by employee and month
            if ledger_account == 5000:
                # Group by employee and month
                transactions = (
                    transactions.groupby("Employee Name 1")[list(MONTH_COLS)]
                    .sum()
                    .reset_index()
                )

                # Replace empty employee names with "(Unknown)"
                transactions["Employee Name 1"] = transactions[
                    "Employee Name 1"
                ].replace("", "(Unknown)")

                for month in MONTH_COLS:
                    transactions[month] = transactions[month].apply(
                        lambda x: "" if x == 0.0 else x
                    )
            else:
                # Sort by month_idx
                transactions = transactions.sort_values(by="month_idx")

            rows = []
            for _, transaction_row in transactions.iterrows():
                if ledger_account == 5000:
                    info_columns = ["Employee Name 1"]
                else:
                    info_columns = [
                        "Employee Name 1",
                        "Spend Category",
                        "Business Reason",
                        "Line Memo",
                        "Operational Transaction",
                    ]
                info_columns = [
                    col_name for col_name in info_columns if col_name in transaction_row
                ]

                row_data = {col_name: None for col_name in _get_column_headers()}
                row_data[_ColumnHeaders.CATEGORY.value.col_name] = (
                    ROW_INDENT_STR
                    + ", ".join(
                        filter(
                            lambda x: x and x.strip(),
                            [transaction_row[c] for c in info_columns],
                        )
                    ).strip()
                )
                for month in MONTH_COLS:
                    row_data[month] = transaction_row[month]
                # if month in row_data:
                #     row_data[month] = transaction_row["Amount"]

                rows.append(row_data)

            transaction_df = pd.DataFrame(rows, columns=_get_column_headers())

            # Add the transaction entries DataFrame as sub-rows to the actuals DataFrame
            if not transaction_df.empty:
                self._insert_subrows(
                    transaction_df, row[_ColumnHeaders.CATEGORY.value.col_name]
                )

        # Get all unhandled ledger accounts
        unhandled_ledger_accounts = (
            set(df_data[account_col_name].unique()) - handled_ledger_accounts
        )

        if unhandled_ledger_accounts:
            warn(
                "Your transaction data contains ledger accounts that are not correctly matched to the budget categories: "
                f"{', '.join(map(str, unhandled_ledger_accounts))}. "
                f"Please reach out to {DEVELOPER_NAME} to add support for these accounts."
            )

    def _add_payroll_data(self):
        """Add payroll data to the account DataFrame."""

        payroll_df = self._df_payroll

        ##################### Process payroll_df #####################
        print(f"There are {len(payroll_df)} payroll entries.")

        # Filter down to relevant grant, where 'Grant' starts with self.account_code
        if self.account_code.startswith("GR"):
            payroll_df = payroll_df[
                payroll_df["Grant"].str.startswith(f"{self.account_code} ")
            ].copy()
        elif self.account_code.startswith("AC"):
            payroll_df = payroll_df[
                payroll_df["Activity"].str.startswith(f"{self.account_code} ")
            ].copy()
        elif self.account_code.startswith("GF"):
            payroll_df = payroll_df[
                payroll_df["Gift"].str.startswith(f"{self.account_code} ")
            ].copy()
        else:
            raise ValueError("Unsupported account code for payroll data.")
        print(f"There are {len(payroll_df)} payroll entries for {self.account_code}.")

        # Add a 'Month' column based on 'Budget Date'
        payroll_df["Month"] = pd.to_datetime(payroll_df["Budget Date"]).dt.month

        # Clean the 'Amount' column
        payroll_df["Amount"] = payroll_df["Amount"].apply(workday_str_amount_to_float)

        # Step 2: Group by Employee and Month
        payroll_summary = (
            payroll_df.groupby(["Employee", "Month"])["Amount"].sum().reset_index()
        )

        # Step 3: Pivot: one row per Employee, one column per month
        # and format the month names
        payroll_pivot = payroll_summary.pivot(
            index="Employee", columns="Month", values="Amount"
        ).fillna(0)
        month_cols = [datetime(1900, m, 1).strftime("%b") for m in range(1, 13)]
        payroll_pivot.columns = [
            datetime(1900, int(m), 1).strftime("%b") for m in payroll_pivot.columns
        ]

        # Convert the data to float
        payroll_pivot = payroll_pivot.astype(float)

        # Make sure all months are there
        for m in month_cols:
            if m not in payroll_pivot.columns:
                payroll_pivot[m] = 0.0

        # Prepare the df for merging with actuals_df
        payroll_pivot[_ColumnHeaders.BUDGET.value.col_name] = ""
        payroll_pivot[_ColumnHeaders.PREV_ACTUALS.value.col_name] = ""
        payroll_pivot[_ColumnHeaders.BALANCE.value.col_name] = ""
        payroll_pivot[_ColumnHeaders.COMMITTED.value.col_name] = ""

        ytd = payroll_pivot[month_cols].sum(axis=1).astype(float)
        month_vals = payroll_pivot[month_cols].astype(float)
        payroll_pivot[month_cols] = month_vals.where(month_vals != 0.0, "")
        payroll_pivot[_ColumnHeaders.ACTUALS_YTD.value.col_name] = ytd

        # Intent employee names a bit
        payroll_pivot[_ColumnHeaders.CATEGORY.value.col_name] = payroll_pivot.index.map(
            lambda x: f"{ROW_INDENT_STR}{x}"
        )
        payroll_pivot = payroll_pivot[_get_column_headers()]
        print(
            f"There is data for {len(payroll_pivot)} employees within these payroll entries."
        )
        if payroll_pivot.empty:
            warn(
                "No payroll data found for this account. "
                "Skipping payroll data addition."
            )
            return

        # Merge with the actuals_df
        self._insert_subrows(payroll_pivot, "5000:Salaries and Wages")

    def _insert_subrows(self, df_subrows, after_ledger_account):
        # Drop columns that are all-NA
        df_subrows = df_subrows.dropna(axis=1, how="all")

        category_indices = self._df.index[
            self._df[_ColumnHeaders.CATEGORY.value.col_name] == after_ledger_account
        ]
        if category_indices.empty:
            error(
                f"Could not find the category ('{after_ledger_account}') in the DataFrame."
            )
        category_index = category_indices[0]

        before = self._df.iloc[: category_index + 1]
        after = self._df.iloc[category_index + 1 :]
        self._df = pd.concat([before, df_subrows, after], ignore_index=True)

    ############################################################################
    ##################### Methods to export to Excel and format ################
    ############################################################################

    def to_excel(self, output_path=None):
        if output_path is None:
            output_path = f"{self.account_code}_{self.year}.xlsx"
        print_color(TermColors.GREEN, f"Creating {output_path}")

        formatted_df = self._df.copy()

        formatted_df.to_excel(output_path, index=False, engine="openpyxl")
        self._format_excel(output_path)

    def _format_excel(self, excel_filename):
        """Format the Excel file with specific styles."""
        wb = load_workbook(excel_filename)
        ws = wb.active
        currency_format = "#,##0.00"

        # Insert Title Row
        ws.insert_rows(1)
        title_text = f"{self.account_code} - Year {self.year} (Generated {datetime.now().strftime('%b %d, %Y')})"
        ws.cell(row=1, column=1).value = title_text

        num_columns = ws.max_column
        ws.merge_cells(start_row=1, start_column=1, end_row=1, end_column=num_columns)

        title_cell = ws.cell(row=1, column=1)
        title_cell.font = Font(size=18, bold=True)
        title_cell.alignment = Alignment(horizontal="center", vertical="center")
        ws.row_dimensions[1].height = 30

        # Define a thin black border
        thin_border = Border(
            left=Side(style="thin", color="000000"),
            right=Side(style="thin", color="000000"),
            top=Side(style="thin", color="000000"),
            bottom=Side(style="thin", color="000000"),
        )

        # === Insert new header row ===
        ws.insert_rows(2)  # Insert a new row 2
        for col, header_text in enumerate(_get_column_headers(), start=1):
            ws.cell(row=3, column=col).value = header_text

        # Add "Actuals" merged across Jan to YTD
        first_actuals_col = _ColumnHeaders.JAN.value.index  # 4
        last_actuals_col = _ColumnHeaders.ACTUALS_YTD.value.index  # 16
        ws.merge_cells(
            start_row=2,
            start_column=first_actuals_col,
            end_row=2,
            end_column=last_actuals_col,
        )
        actuals_cell = ws.cell(row=2, column=first_actuals_col)
        actuals_cell.value = _ColumnHeaders.ACTUALS_YTD.value.col_name
        actuals_cell.alignment = Alignment(horizontal="center", vertical="center")
        actuals_cell.font = Font(bold=True)

        # Apply borders to the merged "Actuals" cells
        for col in range(first_actuals_col, last_actuals_col + 1):
            ws.cell(row=2, column=col).border = thin_border

        # Merge vertically: Category, Budget, Prev Actuals, Balance
        for col in [
            _ColumnHeaders.CATEGORY,
            _ColumnHeaders.BUDGET,
            _ColumnHeaders.PREV_ACTUALS,
            _ColumnHeaders.COMMITTED,
            _ColumnHeaders.BALANCE,
        ]:
            ws.merge_cells(
                start_row=2,
                start_column=col.value.index,
                end_row=3,
                end_column=col.value.index,
            )
            merged_cell = ws.cell(row=2, column=col.value.index)
            merged_cell.value = col.value.col_name
            merged_cell.alignment = Alignment(horizontal="center", vertical="center")
            merged_cell.font = Font(bold=True)

            # Apply borders to the merged cells
            ws.cell(row=2, column=col.value.index).border = thin_border
            ws.cell(row=3, column=col.value.index).border = (
                thin_border  # Also border the lower merged part
            )

        # Apply borders to all header cells in row 3
        for col in range(1, num_columns + 1):
            cell = ws.cell(row=3, column=col)
            cell.alignment = Alignment(horizontal="center", vertical="center")
            cell.font = Font(bold=True)
            cell.border = thin_border

        # Set all header cells to have wrap text
        for col in range(1, num_columns + 1):
            header_cell = ws.cell(row=2, column=col)
            header_cell.alignment = Alignment(
                horizontal="center", vertical="center", wrap_text=True
            )

        # Style the second header row (row 3)
        for col in range(1, num_columns + 1):
            header_cell = ws.cell(row=3, column=col)
            header_cell.alignment = Alignment(horizontal="center", vertical="center")
            header_cell.font = Font(bold=True)

        # Freeze top 3 rows
        ws.freeze_panes = ws["B4"]

        # === Format the data rows ===
        fill_5 = PatternFill(
            start_color="F2F2F2", end_color="F2F2F2", fill_type="solid"
        )
        fill_15 = PatternFill(
            start_color="D9D9D9", end_color="D9D9D9", fill_type="solid"
        )
        fill_total = PatternFill(
            start_color="595959", end_color="595959", fill_type="solid"
        )
        yellow_fill = PatternFill(
            start_color="FFFF00", end_color="FFFF00", fill_type="solid"
        )

        rows_to_group = []
        shade_toggle = True

        for idx, row in enumerate(
            ws.iter_rows(min_row=4, min_col=1), start=4
        ):  # Data now starts at row 4
            category_cell = row[0]
            is_employee = isinstance(
                category_cell.value, str
            ) and category_cell.value.startswith(ROW_INDENT_STR)

            if not is_employee:
                fill = fill_5 if shade_toggle else fill_15
                for cell in row:
                    cell.fill = fill
                shade_toggle = not shade_toggle
            else:
                rows_to_group.append(idx)

            for cell in row[1:]:
                cell.number_format = currency_format

        if rows_to_group:
            # Group the rows with the row above
            ws.sheet_properties.outlinePr.summaryBelow = False

            sorted_rows = sorted(rows_to_group)
            start = sorted_rows[0]
            end = start

            for row in sorted_rows[1:]:
                if row == end + 1:
                    # Continue the block
                    end = row
                else:
                    # End current block, start a new one
                    ws.row_dimensions.group(start, end, outline_level=1, hidden=True)
                    start = end = row

            # Group the last block
            ws.row_dimensions.group(start, end, outline_level=1, hidden=True)

        ws.sheet_format = SheetFormatProperties(outlineLevelRow=1)

        # Set column widths
        ws.column_dimensions[
            get_column_letter(_ColumnHeaders.CATEGORY.value.index)
        ].width = 50
        ws.column_dimensions[
            get_column_letter(_ColumnHeaders.BUDGET.value.index)
        ].width = 13
        ws.column_dimensions[
            get_column_letter(_ColumnHeaders.PREV_ACTUALS.value.index)
        ].width = 13
        ws.column_dimensions[
            get_column_letter(_ColumnHeaders.ACTUALS_YTD.value.index)
        ].width = 13
        ws.column_dimensions[
            get_column_letter(_ColumnHeaders.COMMITTED.value.index)
        ].width = 13
        ws.column_dimensions[
            get_column_letter(_ColumnHeaders.BALANCE.value.index)
        ].width = 13
        for month in range(
            _ColumnHeaders.JAN.value.index, _ColumnHeaders.DEC.value.index + 1
        ):
            ws.column_dimensions[get_column_letter(month)].width = 12

        # Highlight the total row
        total_row = len(ws["A"])
        for cell in ws[total_row]:
            cell.fill = fill_total
            cell.font = Font(bold=True, color="FFFFFF")

        balance_cell = ws.cell(row=total_row, column=_ColumnHeaders.BALANCE.value.index)
        balance_cell.fill = yellow_fill
        balance_cell.font = Font(bold=True, color="000000")

        wb.save(excel_filename)



import time
from datetime import datetime

from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys

from .config import (
    WAIT_TIME_GRANT,
    WAIT_TIME_MENU_BUTTON,
    WAIT_TIME_PERIOD,
    WAIT_TIME_BASE,
    WAIT_TIME_TABLE,
    get_verbose,
)
from .chrome import goto_workday_page
from .table import (
    extract_actuals_from_activity_table,
    extract_actuals_from_grant_table,
    extract_journal_data_from_table,
    extract_payroll_data_from_table,
    extract_transaction_data_from_table,
)
from .yutils import TermColors, print_color, print_progress


################################################################################
############################# Report Functions #################################
################################################################################
def run_grant_report(dr, account, month_year):
    _configure_account_report(dr, "grant", account, month_year)
    _run_report(dr)
    _wait_for_data_table(dr, "div.wd-DataGrid")
    ret = extract_actuals_from_grant_table(dr, month_year)
    print("Done")
    return ret


def run_activity_report(dr, account, month_year):
    _configure_account_report(dr, "activity", account, month_year)
    _run_report(dr)
    _wait_for_data_table(dr, "div.wd-DataGrid")
    ret = extract_actuals_from_activity_table(dr, month_year)
    print("Done")
    return ret


def run_gift_report(dr, account, month_year):
    _configure_account_report(dr, "gift", account, month_year)
    _run_report(dr)
    _wait_for_data_table(dr, "div.wd-DataGrid")
    ret = extract_actuals_from_activity_table(dr, month_year)
    print("Done")
    return ret


def run_journal_report(dr, account, month_year):
    _configure_account_report(dr, "journal", account, month_year)
    _run_report(dr)
    _wait_for_data_table(dr, 'table[data-testid="table"]')
    ret = extract_journal_data_from_table(dr)
    print("Done")
    return ret


def run_transactions_report(dr, account, year):
    _configure_account_report(dr, "transaction", account, datetime(year, 1, 1))
    _run_report(dr, click_filter_first=False)
    _wait_for_data_table(dr, 'table[data-testid="table"]')
    ret = extract_transaction_data_from_table(dr)
    print(f"Done. Extracted {len(ret)} rows.")
    return ret


def run_payroll_report(dr, account, year):
    _configure_account_report(dr, "payroll", account, datetime(year, 1, 1))

    _run_report(dr)
    _wait_for_data_table(dr, 'table[data-testid="table"]')

    ret = extract_payroll_data_from_table(dr)
    print("Done")
    return ret


################################################################################
############################# Internal Helpers #################################
################################################################################


def _wait_for_data_table(dr, selector):
    # Wait for the data table to be present and visible
    WebDriverWait(dr, WAIT_TIME_TABLE).until(
        EC.visibility_of_element_located((By.CSS_SELECTOR, selector))
    )
    print_progress("Data table is visible.")


def _configure_account_report(dr, type, account, month_year):
    if type in ("transaction", "payroll"):
        month_year_str = month_year.strftime("%Y")
    else:
        month_year_str = month_year.strftime("%b %Y")

    print_color(
        TermColors.BLUE,
        f"Running {type} report for {account} {month_year_str}",
        end="\n" if get_verbose() else "",
    )
    goto_workday_page(dr)
    _goto_finance_reports(dr)

    if type == "journal":
        _click_finance_report(dr, "BYU Find Journal Lines")
    elif type == "transaction":
        _click_finance_report(dr, "BYU Grants Transactions")
    elif type == "payroll":
        _click_finance_report(dr, "BYU Biweekly Payroll Detail by Worktag")
    else:
        _click_finance_report(
            dr, f"BYU Budget Vs Actuals Overview by {type.capitalize()}"
        )

    # Set the grant or account in the search input
    if type in ("gift", "journal", "grant", "activity"):
        _set_report_input_box(dr, "Organization", account, sleep=WAIT_TIME_GRANT)
    elif type in ("transaction"):
        _set_report_input_box(dr, "Grant", account, sleep=WAIT_TIME_GRANT)

    # For payroll report, click the 'Exclude Grant' checkbox
    if type == "payroll":
        checkbox_panel = WebDriverWait(dr, 10).until(
            EC.element_to_be_clickable(
                (
                    By.XPATH,
                    "//label[contains(text(), 'Exclude Grant')]/ancestor::li//div[@data-automation-id='checkboxPanel']",
                )
            )
        )
        checkbox_panel.click()

    if type == "transaction":
        _set_date_by_label(dr, "Start Date", 1, 1, month_year.year)
        _set_date_by_label(dr, "End Date", 12, 31, month_year.year)
    elif type == "payroll":
        _set_date_by_label(dr, "Starting Accounting_Date", 1, 1, month_year.year)
        _set_date_by_label(dr, "Ending Accounting_Date", 12, 31, month_year.year)
    else:
        _set_report_input_box(
            dr, "Period", month_year.strftime("%b %Y"), sleep=WAIT_TIME_PERIOD
        )


def _run_report(dr, click_filter_first=True):
    # Click in the Filter Name box before clicking OK
    if click_filter_first:
        filter_name_input = WebDriverWait(dr, 10).until(
            EC.element_to_be_clickable(
                (By.CSS_SELECTOR, 'input[data-automation-id="textInputBox"]')
            )
        )
        dr.execute_script(
            "arguments[0].scrollIntoView({block: 'center'});", filter_name_input
        )
        filter_name_input.click()
        print_progress("Clicked in the Filter Name input box.")

    # Now wait for the OK button to be clickable and click it
    try:
        ok_button = WebDriverWait(dr, WAIT_TIME_BASE).until(
            EC.element_to_be_clickable(
                (
                    By.CSS_SELECTOR,
                    'button[data-automation-id="wd-CommandButton_uic_okButton"][title="OK"]',
                )
            )
        )
        dr.execute_script("arguments[0].scrollIntoView({block: 'center'});", ok_button)
        ok_button.click()
        print_progress("OK button to run the report.")
    except Exception as e:
        print("Could not find or click the OK button:", e)


def _set_date_by_label(dr, label_text, month, day, year):
    _set_report_input_box(
        dr,
        label_text,
        str(month).zfill(2),
        "[@data-automation-id='dateSectionMonth-input']",
    )
    _set_report_input_box(
        dr,
        label_text,
        str(day).zfill(2),
        "[@data-automation-id='dateSectionDay-input']",
    )
    _set_report_input_box(
        dr,
        label_text,
        str(year).zfill(2),
        "[@data-automation-id='dateSectionYear-input']",
    )
    return


def _goto_finance_reports(dr):
    # Click the global navigation button to open the menu
    global_nav_button = WebDriverWait(dr, WAIT_TIME_BASE).until(
        EC.element_to_be_clickable(
            (By.CSS_SELECTOR, 'button[data-automation-id="globalNavButton"]')
        )
    )
    time.sleep(WAIT_TIME_MENU_BUTTON)
    global_nav_button.click()
    print_progress("Menu button clicked.")

    # Click the Menu button
    menu_button = WebDriverWait(dr, WAIT_TIME_BASE).until(
        EC.element_to_be_clickable(
            (By.CSS_SELECTOR, 'button[data-automation-id="globalNavAppsTab"]')
        )
    )
    menu_button.click()
    print_progress("Menu button clicked.")

    # Wait for and click the BYU Finance Dashboard link
    finance_dashboard_link = WebDriverWait(dr, WAIT_TIME_BASE).until(
        EC.element_to_be_clickable(
            (
                By.CSS_SELECTOR,
                'a[data-automation-id="globalNavAppItemLink"][aria-label="BYU Finance Dashboard"]',
            )
        )
    )
    dr.execute_script(
        "arguments[0].scrollIntoView({block: 'center'});", finance_dashboard_link
    )
    WebDriverWait(dr, WAIT_TIME_BASE).until(EC.visibility_of(finance_dashboard_link))
    finance_dashboard_link.click()
    print_progress("BYU Finance Dashboard link clicked.")

    # Wait for and click the BYU Finance Reports tab
    tabs = WebDriverWait(dr, WAIT_TIME_BASE).until(
        EC.presence_of_all_elements_located(
            (
                By.CSS_SELECTOR,
                'li[data-automation-id="tab"] .gwt-Label[data-automation-id="tabLabel"]',
            )
        )
    )
    found = False
    for tab in tabs:
        if tab.text.strip() == "BYU Finance Reports":
            tab.click()
            print_progress("BYU Finance Reports tab clicked.")
            found = True
            break
    if not found:
        print("Could not find the BYU Finance Reports tab.")
    time.sleep(1)

    # Wait for and click the 'More' button before clicking the report
    try:
        # Wait for the 'More' button to be present in the DOM
        more_button = WebDriverWait(dr, WAIT_TIME_BASE).until(
            EC.element_to_be_clickable(
                (By.CSS_SELECTOR, 'div[data-automation-id="wd-MoreLink"]')
            )
        )
        # Wait until the button is visible (not zero size)
        WebDriverWait(dr, WAIT_TIME_BASE).until(
            lambda d: more_button.is_displayed()
            and more_button.size["height"] > 0
            and more_button.size["width"] > 0
        )
        dr.execute_script(
            "arguments[0].scrollIntoView({block: 'center'});", more_button
        )
        # Try to click with JS if normal click fails
        try:
            more_button.click()
        except Exception:
            dr.execute_script("arguments[0].click();", more_button)
        print_progress("'More' button clicked.")
    except Exception as e:
        print("Could not find or click the 'More' button:", e)
    time.sleep(1)


def _click_finance_report(dr, report_name):
    """On the 'BYU Finance Reports' page, click the specified report by its name."""
    report_div = WebDriverWait(dr, WAIT_TIME_BASE).until(
        EC.element_to_be_clickable(
            (
                By.XPATH,
                f"//div[@title='{report_name}']",
            )
        )
    )
    dr.execute_script("arguments[0].scrollIntoView({block: 'center'});", report_div)
    report_div.click()
    print_progress(f"'{report_name}' report clicked.")


def _set_report_input_box(dr, label_name, value, filter="", sleep=0):
    search_path = f"//label[normalize-space(.)='{label_name}']/ancestor::li//input"
    if filter:
        search_path += filter
    input_box = WebDriverWait(dr, WAIT_TIME_BASE).until(
        EC.element_to_be_clickable(
            (
                By.XPATH,
                search_path,
            )
        )
    )
    dr.execute_script("arguments[0].scrollIntoView({block: 'center'});", input_box)
    dr.execute_script("arguments[0].focus();", input_box)
    input_box.clear()
    print_progress(f"Setting {label_name} to {value}")
    input_box.send_keys(value)
    time.sleep(0.5)
    input_box.send_keys(Keys.ENTER)
    time.sleep(sleep)
